# Copyright 2016-2018 PeppyMeter peppy.player@gmail.com
# 
# This file is part of PeppyMeter.
# 
# PeppyMeter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PeppyMeter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PeppyMeter. If not, see <http://www.gnu.org/licenses/>.

import math
import time
import statistics
import logging
import audioop
import threading

from random import uniform
from threading import Thread, RLock
from configfileparser import *

SOURCE_CONSTANT = "constant"
SOURCE_NOISE = "noise"
SOURCE_SAW = "saw"
SOURCE_TRIANGLE = "triangle"
SOURCE_SINE = "sine"
SOURCE_PIPE = "pipe"

MONO_ALGORITHM_MAXIMUM = "maximum"
MONO_ALGORITHM_AVERAGE = "average"

STEREO_ALGORITHM_NEW = "new"
STEREO_ALGORITHM_LOGARITHM = "logarithm"
STEREO_ALGORITHM_AVERAGE = "average"

LEFT = (1, 0)
RIGHT = (0, 1)


class DataSource(object):
    """ Provides methods to generate different types of audio signal. """
    
    lock = RLock()
    
    def __init__(self, c):
        """ Initializer
        
        :param c: configuration dictionary
        """
        self.config = c[DATA_SOURCE]
        self.mono_algorithm = self.config[MONO_ALGORITHM]
        self.stereo_algorithm = self.config[STEREO_ALGORITHM]
        self.ds_type = self.config[TYPE]
        self.const = self.config[VOLUME_CONSTANT]
        self.pipe_name = self.config[PIPE_NAME]
        self.min = self.config[VOLUME_MIN]
        self.max_in_ui = self.config[VOLUME_MAX]
        self.max_in_pipe = self.config[VOLUME_MAX_IN_PIPE]
        
        self.v = 0
        self.step = self.config[STEP]
        self.pipe_size = 4
        self.rng = list(range(int(self.min), int(self.max_in_ui)))
        self.double_rng = self.rng
        self.double_rng.extend(range(int(self.max_in_ui) - 1, int(self.min), -1))
        self.pipe = None
        if self.ds_type == SOURCE_PIPE:
            thread = Thread(target=self.open_pipe)
            thread.start()
        self.previous_left = self.previous_right = self.previous_mono = 0.0
        self.run_flag = True
        self.polling_interval = self.config[POLLING_INTERVAL]
        self.prev_time = None
        self.data = ()
        self.prevdata = (0,0,0)
    
    def open_pipe(self):
        try:            
            logging.debug("opening pipe...")
            self.pipe = os.open(self.pipe_name, os.O_RDONLY | os.O_NONBLOCK)
            logging.debug("pipe opened")
        except:
            logging.debug("Cannot open named pipe: " + self.pipe_name)
            os._exit(0)
        # read previous garbase from the pipe
        try:
            data = os.read(self.pipe, 1000000)
            logging.debug("Read grabase " + str(len(data)))
        except Exception as e:
            logging.debug("Read grabase exception")
            logging.debug(e)

    def start_data_source(self):
        """ Start data source thread. """ 
        #if self.ds_type == SOURCE_PIPE:
        self.run_flag = True
        thread = Thread(target=self.get_data)
        thread.start()
        
    def stop_data_source(self):
        """ Stop data source thread. """ 
               
        self.run_flag = False
    
    def get_current_data(self):
        """ Return current data """

        #if self.ds_type != SOURCE_PIPE:
        #    self.data = self.get_value()
        
        #print("current",self.data) 
        with self.lock:
            return self.data
        
    def get_current_left_channel_data(self):
        """ Return current left channel value """

        #if self.ds_type != SOURCE_PIPE:
        #    self.data = self.get_value()
        #    return self.data[0]
        
        #print("left",self.data) 
        with self.lock:
            if self.data and self.data[0]:
                return self.data[0]
    
    def get_current_right_channel_data(self):
        """ Return current right channel value """
        
        #if self.ds_type != SOURCE_PIPE:
        #    self.data = self.get_value()
        #    return self.data[1]

        #print("right",self.data) 
        with self.lock:
            if self.data and self.data[1]:
                return self.data[1]
        
    def get_current_mono_channel_data(self):
        """ Return current mono value """

        #if self.ds_type != SOURCE_PIPE:
        #    self.data = self.get_value()
        #    return self.data[2]

        #print(threading.currentThread().native_id,"mono",self.data) 
        with self.lock:
            if self.data and self.data[2]:
                return self.data[2]
    
    def get_data(self):
        """ Thread method. """ 
               
        #print("-------start-------------")
        while self.run_flag:
            #self.data = self.get_value()
            self.get_value()
	    #don't sleep on pipes.all the data must be read from pipe	
            #if self.ds_type != SOURCE_PIPE:
            time.sleep(self.polling_interval)
    
    def smooth_val(self, data, prev):
        if data >= prev:
              #return data # goes up immediately floowing the peeks
              return int(prev+100*(data-prev)/120) # go up rapidly, but not immediately
        else:
              return int(prev-100*(prev-data)/450) # go down slowly, better rounding 
              #return data 

    def smooth(self, val):
        #print(len(val),type(val))
        val=[ (self.smooth_val(val[i],self.prevdata[i])) for i in range(3)]
        self.prevdata=val;
        return val 

    def get_value(self):
        """ Get data depending on the data source type. """ 
               
        d = ()
        if self.ds_type == SOURCE_CONSTANT:
            d = self.get_constant_value()
        elif self.ds_type == SOURCE_NOISE:
            d = self.get_noise_value()
        elif self.ds_type == SOURCE_SAW:
            d = self.get_saw_value()
        elif self.ds_type == SOURCE_TRIANGLE:
            d = self.get_triangle_value()
        elif self.ds_type == SOURCE_SINE:
            d = self.get_sine_value()
        elif self.ds_type == SOURCE_PIPE:
	    # male smoothing. for pipes only, With testing signal no smoothing required
            d = self.get_pipe_value()
# when reading the entore stream from mpd.fifo it causes sounf glitches once per hour. so use peppyalsa that cacls everything required
# no smoothing while using peppyalsa, that already does it
            d = self.smooth(d)
            #print(d)  
            #print(threading.currentThread().native_id,"----------",d,self.polling_interval)
        with self.lock:
            self.data = d
        #return d
    
    def get_constant_value(self):
        """ Returns constant value for all channels. The value is defined in the configuration file. """ 
               
        return (self.const, self.const, self.const)
    
    def get_noise_value(self):
        """ Generate random value for all channels. """
        
        new_left = uniform(self.min, self.max_in_ui)
        new_right = uniform(self.min, self.max_in_ui)            
        new_mono = self.get_mono(new_left, new_right)
        
        left = self.get_channel(self.previous_left, new_left)
        right = self.get_channel(self.previous_right, new_right)
        mono = self.get_channel(self.previous_mono, new_mono)
        
        self.previous_left = new_left
        self.previous_right = new_right
        self.previous_mono = new_mono
                
        return (left, right, mono)
    
    def get_saw_value(self):
        """ Generate saw shape signal. """ 
        
        value = self.rng[int(self.v)]      
        s = (value, value, value)
        self.v = (self.v + self.step) % self.max_in_ui
        return s
    
    def get_triangle_value(self):
        """ Generate triangle shape signal. """
        
        value = self.double_rng[self.v]
        t = (value, value, value)
        self.v = (self.v + self.step) % (int(self.max_in_ui * 2 - 1))
        return t
    
    def get_sine_value(self):
        """ Generate sine shape signal. """ 
               
        a = int(self.max_in_ui * ((1 + math.sin(math.radians(-90 + self.v)))/2))
        s = (a, a, a)
        self.v = (self.v + self.step) % 360
        return s
    
    def get_byte_conv2(self,val1,val2):

         signal = val1 + (val2 << 8)
         #print (signal)
         if signal >= 0x8000:
             signal = ~signal & 0x7FFF
             signal = abs(-signal -1)
         #print (signal)
         return (signal)

    #from alsa
    def get_pipe_value_alsa(self):
        """ Get signal from the named pile. """ 
               
        data = None
        left = right = mono = 0.0
        
        if self.pipe == None:
            return (left, right, mono)
        
        try:
            data = os.read(self.pipe, self.pipe_size)
            length = len(data) 
            if length == 0:
                return (self.previous_left, self.previous_right, self.previous_mono)
            #print(data)
            new_left = int(self.max_in_ui * ((data[length - 4] + (data[length - 3] << 8)) / self.max_in_pipe))
            new_right = int(self.max_in_ui * ((data[length - 2] + (data[length - 1] << 8)) / self.max_in_pipe))
            new_mono = self.get_mono(new_left, new_right)
            #print(new_left,new_right,new_mono)
            left = self.get_channel(self.previous_left, new_left)
            right = self.get_channel(self.previous_right, new_right)
            mono = self.get_channel(self.previous_mono, new_mono)
            
            self.previous_left = new_left
            self.previous_right = new_right
            self.previous_mono = new_mono
        except Exception as e:
            logging.debug(e)
            return (0,0,0)
        
        return (self.previous_left, self.previous_right, self.previous_mono)

# when reading the entore stream from mpd.fifo it causes sounf glitches once per hour. so use peppyalsa that cacls everything required
#    def get_pipe_value_from_mpd_not_used(self):
    def get_pipe_value(self):
        """ Get signal from the named pile. """ 
               
        data = None
        left = right = mono = 0.0
        
        if self.pipe == None:
            return (left, right, mono)
        
        #try:
        #----------readsize=int(self.polling_interval*44100*4*8); # should be above the expected size, to read all the data
        readsize=1000#int(self.polling_interval*44100*4*8); # should be above the expected size, to read all the data
        #print("readsize",readsize)
        #readstep=readsize
        roundright=roundleft=0
        #exceptioncount=0
        #totalread=0
        try:
                data = os.read(self.pipe, readsize)
        	#data = os.read(self.pipe, self.pipe_size)
        	# use polling_interval as rounding interval
                length = len(data) 
                #totalread=totalread+length
                #print("----",length,readsize)
                #print(data)	
                if length == 0:
                    #print("zero len")
                    return (0, 0, 0)

                #this is for pipe from alsa (with average/peak calculatio already performed)
                scope = range(0, length, 4)
                for i in scope:
                    tmp_left = data[i] + (data[i+1] << 8)
                    tmp_right =data[i+2] + (data[i+3] << 8)
                    if tmp_left>roundleft:
                        roundleft=tmp_left
                    if tmp_right>roundright:
                        roundright=tmp_right
                    #print(tmp_left,tmp_right)
                #print("A",length,roundleft,roundright)

                #this is for pipe stream from mpd 
                #roundleft  = self.get_pipe_channel(data, LEFT)
                #roundright = self.get_pipe_channel(data, RIGHT)
                #print(length,roundleft,roundright)

                # audioop gives much less CPU usage
                #scope = range(0, length-4, 4)
                #for i in scope:
                #    #if i>roundleft:
                #    #    roundleft=i;
                #    leftsample=self.get_byte_conv2(data[i], data[i+1] )
                #    if leftsample>roundleft:
                #        roundleft=leftsample
                #    rightsample=self.get_byte_conv2(data[i+2], data[i+3] )
                #    if rightsample>roundright:
                #        roundright=rightsample
                #print("A",length,roundleft,roundright)
                #print("A",roundleft,roundright)
                #readsize=readsize-readstep
        except Exception as e:
                #print("exception")
                logging.debug(e)
                return (0,0,0)
 
        
        new_left = int(self.max_in_ui *  roundleft  / self.max_in_pipe)
        new_right = int(self.max_in_ui * roundright  / self.max_in_pipe)
        #print("---",totalread,new_left,new_right)

        #new_left = int(self.max_in_ui *  self.get_byte_conv2(data[length - 4], data[length - 3] ) / self.max_in_pipe)
        #new_right = int(self.max_in_ui * self.get_byte_conv2(data[length - 2], data[length - 1] ) / self.max_in_pipe)

        #new_left = int(self.max_in_ui * ((data[length - 4] + (data[length - 3] << 8)) / self.max_in_pipe))
        #new_right = int(self.max_in_ui * ((data[length - 2] + (data[length - 1] << 8)) / self.max_in_pipe))

        new_mono = self.get_mono(new_left, new_right)
        
        left = self.get_channel(self.previous_left, new_left)
        right = self.get_channel(self.previous_right, new_right)
        mono = self.get_channel(self.previous_mono, new_mono)
        
        self.previous_left = new_left
        self.previous_right = new_right
        self.previous_mono = new_mono
        #except Exception as e:
        #    #time.sleep(self.polling_interval)
        #    print("exception")
        #   logging.debug(e)
        #    return (0, 0, 0)
        
        #print (self.previous_left, self.previous_right, self.previous_mono)
        return (self.previous_left, self.previous_right, self.previous_mono)
    
    def get_pipe_channel(self, data, channel):
        """ Retrieve data for particular channel """
        
        
        ch = audioop.tomono(data, 2, channel[0], channel[1])
        #mul = audioop.mul(ch, 2, n)
        #ch_signal = audioop.rms(mul, 2)       
        #print ("RMS",(self.max_in_ui * audioop.rms(ch, 2)   ) / self.max_in_pipe)
        #print ("RMS",( audioop.rms(ch, 2)   ) )
        #min_max1= audioop.minmax(ch, 2) #(0,0)#   (self.max_in_ui * audioop.minmax(ch, 2)) / self.max_in_pipe
        #print ("MINMAX",min_max1)
        #v = int(self.max * (ch_signal / self.k))
        v = audioop.maxpp(ch, 2) # 65535 full scale, It's peak to peak
        #print ("MAX",self.max_in_ui * v/self.max_in_pipe)
        #print ("MAX", v)
        
        return v

    def get_mono(self, left, right):
        """ Create mono signal from stereo using particular algorithm """ 
                      
        if self.mono_algorithm == MONO_ALGORITHM_MAXIMUM:
            mono = max(left, right)
        elif self.mono_algorithm == MONO_ALGORITHM_AVERAGE:
            mono = statistics.mean([left, right])
        return mono
    
    def get_channel(self, previous_value, new_value):
        """ Prepares signal value depending on the previous one and algorithm. """ 
               
        if self.stereo_algorithm == STEREO_ALGORITHM_NEW:
            channel_value = new_value
        elif self.stereo_algorithm == STEREO_ALGORITHM_LOGARITHM:
            if previous_value == 0.0:
                channel_value = 0.0
            else:
                channel_value = 20 * math.log10(new_value/previous_value)
            if channel_value < -20:
                channel_value = -20
            if channel_value > 3:
                channel_value = 3
            channel_value = (channel_value + 20) * (100/23)
        elif self.stereo_algorithm == STEREO_ALGORITHM_AVERAGE:
            channel_value = statistics.mean([previous_value, new_value])
                
        return channel_value
        
